# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from pyramid.config import Configurator
from pyramid.static import static_view

from dokang.utils import load_doc_sets


def make_app(global_settings, **settings):
    """Set up and return the WSGI application."""
    settings = load_doc_sets(settings)
    config = Configurator(settings=settings)

    # Third-party includes
    config.include('pyramid_chameleon')

    # Routes and views for static and home page
    config.add_static_view('static', 'static')
    config.add_route('search', '/')
    config.add_view('.views.search', route_name='search', renderer='templates/search.pt')

    # OpenSearch
    config.add_route('opensearch', '/opensearch.xml')
    config.add_view('.views.opensearch', route_name='opensearch')

    # Upload endpoint
    config.add_route('doc_upload', '/upload')
    config.add_view('.views.doc_upload', route_name='doc_upload')

    # A catch-all route that serves documentation from the root of
    # Dokang.
    uploaded_docs_dir = settings.get('dokang.uploaded_docs.dir')
    static_doc_view = static_view(uploaded_docs_dir, use_subpath=True)
    config.add_route('catch_all_doc_view', '/*subpath')
    config.add_view(static_doc_view, route_name='catch_all_doc_view')

    return config.make_wsgi_app()



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import unicode_literals
import base64
import itertools
import os
import shutil
import tempfile
import zipfile

from pyramid.httpexceptions import HTTPMovedPermanently, HTTPMethodNotAllowed, HTTPForbidden, HTTPBadRequest
from pyramid.renderers import get_renderer
from pyramid.response import Response

import wtforms
from wtforms import validators as wtvalidators

from dokang import api
from dokang import utils


def get_hit_limit(settings):
    if 'dokang.hit_limit' not in settings:
        return None
    hit_limit = int(settings['dokang.hit_limit'])
    if hit_limit == 0:
        return None
    return hit_limit


def search(request):
    settings = request.registry.settings
    doc_sets = settings['dokang.doc_sets']
    hit_limit = get_hit_limit(settings)
    raw_query = request.GET.get('query')
    only_doc_set = request.GET.get('doc_set')
    if raw_query:
        query = raw_query
        if only_doc_set:
            query += ' set:%s' % only_doc_set
        index_path = settings['dokang.index_path']
        hits = list(api.search(index_path, query, limit=hit_limit))
        for hit in hits:
            hit['doc_set_title'] = doc_sets[hit['set']]['title']
    else:
        hits = None

    sorted_doc_sets = sorted(doc_sets.values(), key=lambda d: d['title'].lower())
    return {
        'api': TemplateApi(request),
        'query': raw_query,
        'only_doc_set': only_doc_set,
        'doc_sets': [
            (k.upper(), list(v))
            for k, v in itertools.groupby(sorted_doc_sets, key=lambda d: d['title'][0].lower())
        ],
        'hits': hits
    }


def opensearch(request):
    """Return OpenSearch description file."""
    settings = request.registry.settings
    params = {
        'name': settings['dokang.opensearch.name'],
        'description': settings['dokang.opensearch.description'],
        'favicon': request.static_url('dokang:static/img/favicon.ico'),
        'search_url': request.route_url('search'),
    }
    return Response(
        body="""<?xml version="1.0" encoding="UTF-8" ?>
<OpenSearchDescription xmlns="http://a9.com/-/spec/opensearch/1.1/">
  <ShortName>%(name)s</ShortName>
  <Description>%(description)s</Description>
  <Image>%(favicon)s</Image>
  <InputEncoding>UTF-8</InputEncoding>
  <OutputEncoding>UTF-8</OutputEncoding>
  <Url type="text/html" template="%(search_url)s?query={searchTerms}"/>a
</OpenSearchDescription>""" % params,
        content_type=b'application/opensearchdescription+xml'
    )


class TemplateApi(object):
    """
    Provide a master template and various information and utilities that can be used in any template.

    Not that we really need that for a single template but, well, that's what I usually do...
    """

    def __init__(self, request):
        self.request = request
        self.layout = get_renderer('templates/layout.pt').implementation()

    def route_url(self, route_name, *elements, **kw):
        return self.request.route_url(route_name, *elements, **kw)

    def hit_url(self, hit):
        return self.doc_url(hit['set'], hit['path'])

    def doc_url(self, doc_set_id, path=''):
        # We suppose that the id of the document set is used to create
        # the upload directory. This is what `utils.doc_set()` does.
        # This assumption simplifies the code here.
        subpath = os.path.join(doc_set_id, path)
        return self.route_url('catch_all_doc_view', subpath=subpath)

    def static_url(self, path, **kw):
        if ':' not in path:
            path = 'dokang:%s' % path
        return self.request.static_url(path, **kw)


class DocUploadForm(wtforms.Form):
    name = wtforms.StringField('Package Name', validators=[wtvalidators.DataRequired('No package name given')])
    content = wtforms.FileField('Documentation Zip')

    @staticmethod
    def validate_content(form, field):
        if field.data.bufsize > 100 * 1024 * 1024:
            raise wtvalidators.ValidationError('ZIP file is too large')

        data_file = field.data.file
        if not zipfile.is_zipfile(data_file):
            raise wtvalidators.ValidationError('ZIP file is not a zipfile')

        zip_file = zipfile.ZipFile(data_file)
        members = zip_file.namelist()
        if 'index.html' not in members:
            raise wtvalidators.ValidationError('Top-level "index.html" missing in the ZIP file')

        base_dir = tempfile.mkdtemp()
        try:
            for name in members:
                if not os.path.normpath(os.path.join(base_dir, name)).startswith(base_dir):
                    raise wtvalidators.ValidationError('Invalid path name %s in the ZIP file', name)
        finally:
            os.rmdir(base_dir)

        data_file.seek(0)


def doc_upload(request):  # Route is not activated when dokang.uploaded_docs.dir is not set
    settings = request.registry.settings
    doc_dir = settings['dokang.uploaded_docs.dir']
    token = settings.get('dokang.uploaded_docs.token')

    bad_auth = (
        token is None or
        request.authorization is None or
        request.authorization[0] != 'Basic' or
        base64.b64decode(request.authorization[1]) != 'dokang:{0}'.format(token)
    )
    if bad_auth:
        raise HTTPForbidden()

    if not request.POST:
        raise HTTPMethodNotAllowed()

    if request.POST.get(':action', '--no-action--') != 'doc_upload':
        raise HTTPBadRequest('Only doc_upload action is supported.')

    form = DocUploadForm(request.POST)
    if not form.validate():
        raise HTTPBadRequest(form.errors)

    project = form.data['name']
    project_dir = os.path.join(doc_dir, project)

    zip_file = zipfile.ZipFile(form.data['content'].file)
    if os.path.exists(project_dir):
        shutil.rmtree(project_dir)
    zip_file.extractall(project_dir)

    # index new doc set
    index_path = settings['dokang.index_path']
    api.index_document_set(index_path, utils.doc_set(settings, project), force=False)

    # update doc sets
    settings['dokang.doc_sets'][project] = utils.doc_set(settings, project)

    return HTTPMovedPermanently(request.route_url('catch_all_doc_view', subpath=project))



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import unicode_literals

import logging
import os

from bs4 import BeautifulSoup

from dokang.harvesters.base import Harvester


logger = logging.getLogger(__name__)


class SphinxHarvester(Harvester):
    """Harvest content from the HTML rendered version of a Sphinx-based
    set of documents.

    We look at the rendered HTML and not the source files to avoid
    wrongly indexing files included with the ``include`` directive.
    """

    def harvest_file(self, path):
        with open(path, 'rb') as fp:
            html = fp.read()
        soup = BeautifulSoup(html, 'html.parser')
        title, content = self._retrieve_title_and_content(soup)
        # Sphinx adds this character after each heading.
        title = title.replace("¶", "")
        content = content.replace("¶", "")
        title = title.strip()
        content = content.strip()
        return {
            'title': title,
            'content': content,
            'kind': 'HTML',
        }

    def _retrieve_title_and_content(self, soup):
        # A 'div class="body"' element is generated by the default
        # theme. I am not sure about other themes, though. See
        # ReadTheDocsSphinxHarvester for a custom harvester.
        content_block = soup.find(class_='body')
        # We prefer not to use the <title> element because Sphinx
        # appends the title of the project to the title of the page.
        title = content_block.find('h1').get_text()
        content = content_block.get_text()
        return title, content


class ReadTheDocsSphinxHarvester(SphinxHarvester):
    """Harvest content from the HTML rendered version of a Sphinx-based
    set of documents that uses the "Read The Docs" theme.

    The "Read The Docs" theme does not generate the ``<div>`` that we
    look for in the super class. We have to look for a different one.
    """

    def _retrieve_title_and_content(self, soup):
        content_block = soup.find(role='main')
        title = content_block.find('h1').get_text()
        content = content_block.get_text()
        return title, content


def sphinx_config(harvester=SphinxHarvester, include=None, exclude=None, **extensions):
    """
    Return a configuration that is suitable for a Sphinx-based documentation.

    If the documentation uses "Read The Docs" theme, you should rather use ``sphinx_rtd_config``.
    """
    # Exclude automatically generated HTML files such as 'search.html'
    if include is None:
        include = ('_download', )
    if exclude is None:
        exclude = (
            '^genindex.html$',
            '^objects.inv$',
            '^search.html$',
            '^searchindex.js$',
            '%s?_.*' % os.path.sep
        )
    config = {
        'include': include,
        'exclude': exclude,
        'html': harvester,
        'htm': harvester,
    }
    config.update(extensions)
    return config


def sphinx_rtd_config(harvester=ReadTheDocsSphinxHarvester, include=None, exclude=None, **extensions):
    """Return a configuration that is suitable for a Sphinx-based documentation that uses the ReadTheDocs theme."""
    return sphinx_config(harvester=harvester, include=include, exclude=exclude, **extensions)



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import unicode_literals

import logging
import os

from dokang.backends import whoosh
from dokang import harvesters


logger = logging.getLogger(__name__)


# This API is very dependent on the Whoosh backend. This is not
# entirely unrelated to the fact that Whoosh is the only backend.
# There is a plan for this API, though. The plan is not to have a plan
# until we need one.


def initialize_index(index_path):
    """Initialize the index."""
    indexer = whoosh.WhooshIndexer(index_path)
    indexer.initialize()


def index_document_set(index_path, info, force=False):
    """Index or reindex the document set.

    ``info`` should be a single item of the ``DOC_SETS`` configuration
    dictionary.

    If ``force`` is provided, all documents are reindexed again, even
    if they were not modified.
    """
    indexer = whoosh.WhooshIndexer(index_path)
    searcher = whoosh.WhooshSearcher(index_path)
    hashes = searcher.get_hashes()

    doc_set = info['id']
    logger.info('Indexing doc set "%s"...', doc_set)

    # Unindex documents that have been deleted from the document set.
    to_be_deleted = []
    for relative_path in hashes[doc_set].keys():
        path = os.path.join(info['path'], relative_path)
        if not os.path.exists(path):
            logger.debug('Marking indexed document "%s" for deletion.', relative_path)
            to_be_deleted.append(relative_path)
    if to_be_deleted:
        indexer.delete_documents(doc_set, to_be_deleted)

    # Index or update all documents, or ignore them if they did not
    # change.
    documents = harvesters.harvest_set(
        info['path'],
        doc_set,
        info['harvester'],
        hashes.get(doc_set, {}),
        force)
    indexer.index_documents(documents)


def clear_document_set(index_path, doc_set):
    """Remove the whole document set from the index."""
    indexer = whoosh.WhooshIndexer(index_path)
    logger.info('Clearing document set "%s"', doc_set)
    indexer.clear_set(doc_set)


def search(index_path, query, limit=None):
    """Return hits matching the query, as an iterator."""
    searcher = whoosh.WhooshSearcher(index_path)
    return searcher.search(query, limit=limit)



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.
"""Index and search backends for Whoosh."""

from __future__ import absolute_import

from collections import defaultdict
import os
import shutil

from whoosh.fields import ID, Schema, STORED, TEXT
from whoosh.index import create_in, open_dir
from whoosh.qparser import MultifieldParser
from whoosh.query import And, Or, Term
from whoosh.writing import AsyncWriter


class WhooshIndexer(object):
    """Encapsulate indexation through Whoosh."""

    def __init__(self, index_path):
        self.index_path = index_path

    def initialize(self):
        """Initialize the index.

        If an index already exists, it is deleted and recreated from
        scratch.
        """
        schema = Schema(
            uid=ID(stored=False, unique=True),
            path=ID(stored=True),
            set=ID(stored=True),
            hash=STORED,  # not searchable
            title=TEXT(stored=True),
            content=TEXT(stored=False),
            kind=TEXT(stored=True),
        )
        if os.path.exists(self.index_path):
            shutil.rmtree(self.index_path)
        os.mkdir(self.index_path)
        create_in(self.index_path, schema)

    def clear_set(self, doc_set):
        """Remove all documents of this set from the index."""
        index = open_dir(self.index_path)
        index.delete_by_term('set', doc_set)

    def index_documents(self, documents):
        """Add or update documents in the index."""
        index = open_dir(self.index_path)
        writer = AsyncWriter(index)
        needs_commit = False
        for document in documents:
            needs_commit = True
            writer.update_document(
                uid=':'.join((document['set'], document['path'])),
                path=document['path'],
                set=document['set'],
                hash=document['hash'],
                title=document['title'],
                content=document['content'],
                kind=document['kind'],
            )
        if needs_commit:
            writer.commit()

    def delete_documents(self, doc_set, paths):
        """Delete documents from the index."""
        index = open_dir(self.index_path)
        writer = AsyncWriter(index)
        query = And([
            Term('set', doc_set),
            Or([Term('path', path) for path in paths])
        ])
        writer.delete_by_query(query)
        writer.commit()


class WhooshSearcher(object):
    """Encapsulate search through Whoosh."""

    def __init__(self, index_path):
        self.index = open_dir(index_path)

    def get_hashes(self):
        """Return the hash of each indexed document."""
        # It is not as heavy as it seems.
        hashes = defaultdict(dict)
        with self.index.searcher() as searcher:
            for doc in searcher.all_stored_fields():
                hashes[doc['set']][doc['path']] = doc['hash']
        return hashes

    def search(self, query_string, limit=None):
        """Search the query string in the index."""
        parser = MultifieldParser(('title', 'content'), self.index.schema)
        query = parser.parse(query_string)
        with self.index.searcher() as searcher:
            for hit in searcher.search(query, limit=limit):
                yield {
                    'path': hit['path'],
                    'title': hit['title'],
                    'set': hit['set'],
                    'kind': hit['kind'],
                }



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.
"""Python 2 compatibility layer."""

from __future__ import print_function

import os
import sys

PY2 = sys.version[0] == '2'

# pylint: disable=redefined-builtin, invalid-name, import-error, unused-import, no-name-in-module
if PY2:
    from ConfigParser import SafeConfigParser as ConfigParser

    def print_to_stdout(unicode_string):
        print(unicode_string.encode(sys.stdout.encoding))
else:
    from configparser import ConfigParser

    def print_to_stdout(string):
        string += os.linesep
        sys.stdout.buffer.write(string.encode(sys.stdout.encoding))



import os
import re

from setuptools import setup, find_packages

# Do not try to import the package to get its version.
_version_file = open(os.path.join(os.path.dirname(__file__), 'dokang', 'version.py'))
VERSION = re.compile(r"^VERSION = '(.*?)'", re.S).match(_version_file.read()).group(1)


def read(filename):
    with open(filename) as fp:
        return fp.read().strip()


setup(
    name='Dokang',
    version=VERSION,
    description="Lightweight web documentation repository with a search engine",
    long_description='%s\n\n%s' % (read('README.rst'), read('CHANGES.txt')),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Text Processing :: Markup :: HTML",
    ],
    author="Polyconseil",
    author_email="opensource+dokang@polyconseil.fr",
    url='https://dokang.readthedocs.org/',
    keywords='documentation repository search engine',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'beautifulsoup4==4.4.0',
        'Chameleon==2.22',
        'pyramid==1.5.7',
        'pyramid_chameleon==0.3',
        'Whoosh==2.7.0',
        'WTForms==2.0.2',
    ],
    tests_require=[l for l in read('requirements_dev.txt').splitlines() if not l.startswith(('-', '#'))],
    test_suite='tests',
    entry_points={
        'paste.app_factory': [
            'main=dokang.app:make_app',
        ],
        'console_scripts': [
            'dokang=dokang.cli:main',
        ],
    },
)



# -*- coding: utf-8 -*-
# Copyright (c) Polyconseil SAS. All rights reserved.

from __future__ import unicode_literals

from bs4 import BeautifulSoup

from dokang.harvesters.base import Harvester


class HtmlHarvester(Harvester):
    """Harvest content from HTML files."""

    def harvest_file(self, path):
        with open(path, 'rb') as fp:
            html = fp.read()
        soup = BeautifulSoup(html, 'html.parser')
        title, content = self._retrieve_title_and_content(soup)
        return {
            'title': title,
            'content': content,
            'kind': 'HTML',
        }

    @staticmethod
    def _retrieve_title_and_content(soup):
        title = soup.title.string.strip()
        content = soup.find('body').get_text().strip()
        return title, content


def html_config(harvester=HtmlHarvester, include=None, exclude=None, **extensions):
    """Return a configuration that is suitable for an HTML document set."""
    config = {
        'include': include,
        'exclude': exclude,
        'html': harvester,
        'htm': harvester,
    }
    config.update(extensions)
    return config
